/*
 * ============================================================================
 * Licensed Materials - Property of IBM
 * Project  Zero
 *
 * (C) Copyright IBM Corp. 2007,2008  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 * ============================================================================
 */
if(!dojo._hasResource["zero.ide.data.ideFilePathStore"]){ //_hasResource checks added by build. Do not use _hasResource directly in your code.
dojo._hasResource["zero.ide.data.ideFilePathStore"] = true;
dojo.provide("zero.ide.data.ideFilePathStore");

dojo.require("dojo.data.util.filter");
dojo.require("dojo.data.util.simpleFetch");
dojo.require("dojo.date.stamp");

dojo.declare("zero.ide.data.ideFilePathStore", null,{
	constructor: function(/* Object */ keywordParameters){
	
		this._arrayOfAllItems = [];
		this._arrayOfTopLevelItems = [];
		this._loadFinished = false;
		this._jsonFileUrl = keywordParameters.url;
		this._jsonData = keywordParameters.data;
		this._datatypeMap = keywordParameters.typeMap || {};
		if(!this._datatypeMap['Date']){
			this._datatypeMap['Date'] = {
											type: Date,
											deserialize: function(value){
												return dojo.date.stamp.fromISOString(value);
											}
										};
		}
		this._features = {'dojo.data.api.Read':true, 'dojo.data.api.Identity':true};
		this._itemsByIdentity = null;
		this._storeRefPropName = "_S";  // Default name for the store reference to attach to every item.
		this._itemNumPropName = "_0"; // Default Item Id for isItem to attach to every item.
		this._rootItemPropName = "_RI"; // Default Item Id for isItem to attach to every item.
		this._loadInProgress = false;	//Got to track the initial load to prevent duelling loads of the dataset.
		this._queuedFetches = [];
		this.resourceString = keywordParameters.resourceString;
		
	},
	
	url: "",

	setUrl: function(newurl, resourceString) {
      //if(zlog.isDebug())console.debug("ideFilePathStore setUrl called newurl ="+newurl);
		if (this.url !== newurl || this.resourceString !== resourceString) {
		   this._arrayOfAllItems = [];
		   this._arrayOfTopLevelItems = [];
		   this._loadFinished = false;
		}		
	    this.resourceString = resourceString;
		this.url = newurl;
		this._jsonFileUrl = newurl;
		//this.resultData = null;
	},


	_assertIsItem: function(/* item */ item){
        //if(zlog.isDebug())console.debug("ideFilePathStore _assertIsItem called ");		

		if(!this.isItem(item)){ 
			throw new Error("zero.ide.data.ideFilePathStore: Invalid item argument.");
		}
	},

	_assertIsAttribute: function(/* attribute-name-string */ attribute){
        //if(zlog.isDebug())console.debug("ideFilePathStore _assertIsAttribute called ");		

		if(typeof attribute !== "string"){ 
			throw new Error("zero.ide.data.ideFilePathStore: Invalid attribute argument.");
		}
	},

	getValue: function(	item, attribute, defaultValue){
        //if(zlog.isDebug())console.debug("ideFilePathStore getValue called attribute ="+attribute);			

		var values = this.getValues(item, attribute);
				
		return (values.length > 0)?values[0]:defaultValue; // mixed
	},

	getValues: function(item, attribute){
        //if(zlog.isDebug())console.debug("ideFilePathStore getValues called ");		

		this._assertIsItem(item);
		this._assertIsAttribute(attribute);
		return item[attribute] || []; // Array
	},

	getAttributes: function(item){
        //if(zlog.isDebug())console.debug("ideFilePathStore getAttributes called ");		

		this._assertIsItem(item);
		var attributes = [];
		for(var key in item){
			if((key !== this._storeRefPropName) && (key !== this._itemNumPropName) && (key !== this._rootItemPropName)){
				attributes.push(key);
			}
		}
		return attributes; // Array
	},

	hasAttribute: function(	item, attribute) {
        //if(zlog.isDebug())console.debug("ideFilePathStore hasAttribute called ");		

		return this.getValues(item, attribute).length > 0;
	},

	containsValue: function(item, attribute, value){
        //if(zlog.isDebug())console.debug("ideFilePathStore containsValue called ");		

		var regexp = undefined;
		if(typeof value === "string"){
			//replace any single backslash with an escaped backslash
			var escapedValue = value.replace(/\\/g,"\\\\");
			
			regexp = dojo.data.util.filter.patternToRegExp(escapedValue, false);
			//if(zlog.isLog())console.log("comparing:"+value);
			//if(zlog.isLog())console.log("with:"+regexp);
		}
		return this._containsValue(item, attribute, value, regexp); //boolean.
	},

	_containsValue: function(	item, attribute, value, regexp){
        //if(zlog.isDebug())console.debug("ideFilePathStore _containsValue called ");		

		return dojo.some(this.getValues(item, attribute), function(possibleValue){

			if(possibleValue !== null && !dojo.isObject(possibleValue) && regexp){
				if(possibleValue.toString().match(regexp)){
					return true; // Boolean
				}
			}else if(value === possibleValue){
				return true; // Boolean
			}
		});
	},

	isItem: function(/* anything */ something){
        //if(zlog.isDebug())console.debug("ideFilePathStore isItem called ");		
		
		if(something && something[this._storeRefPropName] === this){
			if(this._arrayOfAllItems[something[this._itemNumPropName]] === something){
				return true;
			}
		}
		return false; // Boolean
	},

	isItemLoaded: function(/* anything */ something){
        //if(zlog.isDebug())console.debug("ideFilePathStore isitemLoaded called ");		
		
		return this.isItem(something); //boolean
	},

	loadItem: function(/* object */ keywordArgs){
        //if(zlog.isDebug())console.debug("ideFilePathStore loadItem called ");		
		
		this._assertIsItem(keywordArgs.item);
	},

	getFeatures: function(){
        //if(zlog.isDebug())console.debug("ideFilePathStore getFeatures called ");		
		
		return this._features; //Object
	},

	getLabel: function(/* item */ item){
        //if(zlog.isDebug())console.debug("ideFilePathStore getLabel called ");		

		if(this._labelAttr && this.isItem(item)){
			return this.getValue(item,this._labelAttr); //String
		}
		return undefined; //undefined
	},

	getLabelAttributes: function(item){
        //if(zlog.isDebug())console.debug("ideFilePathStore getLabelAttributes called ");		

		if(this._labelAttr){
			return [this._labelAttr]; //array
		}
		return null; //null
	},

	_fetchItems: function(	keywordArgs, findCallback, errorCallback){
        //if(zlog.isDebug())console.debug("ideFilePathStore _fetchItems called ");		
		
		var self = this;
		var filter = function(requestArgs, arrayOfItems){
			var items = [];
			if(requestArgs.query){
				var ignoreCase = requestArgs.queryOptions ? requestArgs.queryOptions.ignoreCase : false; 

				var regexpList = {};
				for(var key in requestArgs.query){
					var value = requestArgs.query[key];
					if(typeof value === "string"){						
						//replace any single backslash with an escaped backslash
						var escapedValue = value.replace(/\\/g,"\\\\");	
						//if(zlog.isLog())console.log("comparing:"+value);											
						regexpList[key] = dojo.data.util.filter.patternToRegExp(escapedValue, ignoreCase);
						//if(zlog.isLog())console.log("with:"+regexpList[key]);
					}
				}

				for(var i = 0; i < arrayOfItems.length; ++i){
					var match = true;
					var candidateItem = arrayOfItems[i];
					if(candidateItem === null){
						match = false;
					}else{
						for(var key in requestArgs.query) {
							var value = requestArgs.query[key];
							if (!self._containsValue(candidateItem, key, value, regexpList[key])){
								match = false;
							}
						}
					}
					if(match){
						items.push(candidateItem);
					}
				}
				findCallback(items, requestArgs);
			}else{
				for(var i = 0; i < arrayOfItems.length; ++i){
					var item = arrayOfItems[i];
					if(item !== null){
						items.push(item);
					}
				}
				findCallback(items, requestArgs);
			}
		};

		if(this._loadFinished){
          //if(zlog.isDebug())console.debug("ideFilePathStore _fetchItems in if loadFinished  ");		

			filter(keywordArgs, this._getItemsArray(keywordArgs.queryOptions));
		}else{
          //if(zlog.isDebug())console.debug("ideFilePathStore _fetchItems in else not loadFinished  ");		
 
			if(this._jsonFileUrl){
				//If fetches come in before the loading has finished, but while
				//a load is in progress, we have to defer the fetching to be 
				//invoked in the callback.
				if(this._loadInProgress){
					this._queuedFetches.push({args: keywordArgs, filter: filter});
				}else{
					this._loadInProgress = true;
					var getArgs = {
							url: self._jsonFileUrl, 
							handleAs: "json-comment-optional",
							load: function(data){
								try{
									newData = {identifier:'path',
									           label:'path',
											   items: data};
									self._getItemsFromLoadedData(newData);
									self._loadFinished = true;
									self._loadInProgress = false;
									filter(keywordArgs, self._getItemsArray(keywordArgs.queryOptions));
									self._handleQueuedFetches();
								}catch(e){
									if(zlog.isDebug())console.debug(e);		
									self._loadFinished = true;
									self._loadInProgress = false;
									errorCallback(e, keywordArgs);
								}
							},
							error: function(error){
								self._loadInProgress = false;
								errorCallback(error, keywordArgs);
							}
						};
					dojo.xhrGet(getArgs);
				}
			}else if(this._jsonData){
				try{
          //if(zlog.isDebug())console.debug("ideFilePathStore _fetchItems in else if jsonData ");		
					
					this._loadFinished = true;
					this._getItemsFromLoadedData(this._jsonData);
					this._jsonData = null;
					filter(keywordArgs, this._getItemsArray(keywordArgs.queryOptions));
				}catch(e){
					errorCallback(e, keywordArgs);
				}
			}else{
				errorCallback(new Error("zero.ide.data.ideFilePathStore: No JSON source data was provided as either URL or a nested Javascript object."), keywordArgs);
			}
		}
	},

	_handleQueuedFetches: function(){
      //if(zlog.isDebug())console.debug("ideFilePathStore _handleQueuedFetches called ");		
		
		if (this._queuedFetches.length > 0) {
			for(var i = 0; i < this._queuedFetches.length; i++){
				var fData = this._queuedFetches[i];
				var delayedQuery = fData.args;
				var delayedFilter = fData.filter;
				if(delayedFilter){
					delayedFilter(delayedQuery, this._getItemsArray(delayedQuery.queryOptions)); 
				}else{
					this.fetchItemByIdentity(delayedQuery);
				}
			}
			this._queuedFetches = [];
		}
	},

	_getItemsArray: function(queryOptions){
      //if(zlog.isDebug())console.debug("ideFilePathStore _getItemsArray called ");		
		
		if(queryOptions && queryOptions.deep) {
			return this._arrayOfAllItems; 
		}
		return this._arrayOfTopLevelItems;
	},

	close: function(/*dojo.data.api.Request || keywordArgs || null */ request){
      //if(zlog.isDebug())console.debug("ideFilePathStore close called ");		

	},

	_getItemsFromLoadedData: function(/* Object */ dataObject){
        //if(zlog.isDebug())console.debug("ideFilePathStore _getItemsFromLoadedData called ");				
		
		// First, we define a couple little utility functions...
		
		function valueIsAnItem(aValue){
      //if(zlog.isDebug())console.debug("ideFilePathStore _getItemsFromLoadedData valueIsAnItem called ");		

			var isItem = (
				(aValue != null) &&
				(typeof aValue == "object") &&
				(!dojo.isArray(aValue)) &&
				(!dojo.isFunction(aValue)) &&
				(aValue.constructor == Object) &&
				(typeof aValue._reference == "undefined") && 
				(typeof aValue._type == "undefined") && 
				(typeof aValue._value == "undefined")
			);
			return isItem;
		}
		
		var self = this;
		function addItemAndSubItemsToArrayOfAllItems(/* Item */ anItem){
      //if(zlog.isDebug())console.debug("ideFilePathStore _getItemsFromLoadedData addItemAndSubItemsToArrayOfAllItems called ");		
      //if(zlog.isDebug())console.debug(anItem["directory"]);
			//self._arrayOfAllItems.push(anItem);
		    if(anItem["children"]){
				delete anItem["children"];
			}
			if (anItem["directory"]) {
			    self._arrayOfAllItems.push(anItem);
				//addItemAndSubItemsToArrayOfAllItems(anItem);
			} 
		}
		

		//this._labelAttr = dataObject.label;
		this._labelAttr = "path";

		// We need to do some transformations to convert the data structure
		// that we read from the file into a format that will be convenient
		// to work with in memory.

		// Step 1: Walk through the object hierarchy and build a list of all items
		var i;
		var item;
		this._arrayOfAllItems = [];
		this._arrayOfTopLevelItems = dataObject.items;
		
		for(i = 0; i < this._arrayOfTopLevelItems.length; ++i){
			item = this._arrayOfTopLevelItems[i];
			
			addItemAndSubItemsToArrayOfAllItems(item);
			item[this._rootItemPropName]=true;
		}		
		
		// Step 2: Walk through all the attribute values of all the items, 
		// and replace single values with arrays.  For example, we change this:
		//		{ name:'Miss Piggy', pets:'Foo-Foo'}
		// into this:
		//		{ name:['Miss Piggy'], pets:['Foo-Foo']}
		// 
		// We also store the attribute names so we can validate our store  
		// reference and item id special properties for the O(1) isItem
		var allAttributeNames = {};
		var key;
		
		for(i = 0; i < this._arrayOfAllItems.length; ++i){
			item = this._arrayOfAllItems[i];			
			
			for(key in item){
				if (key !== this._rootItemPropName)
				{
					var value = item[key];
					
					if(value !== null){
						if(!dojo.isArray(value)){
							var newArray = [];
							newArray[0] = value;
							item[key] = newArray;
						}
					}else{
						item[key] = [null];
					}
				}
				allAttributeNames[key]=key;
			}			
		}
		
		// Step 3: Build unique property names to use for the _storeRefPropName and _itemNumPropName
		// This should go really fast, it will generally never even run the loop.
		while(allAttributeNames[this._storeRefPropName]){
			this._storeRefPropName += "_";
		}
		while(allAttributeNames[this._itemNumPropName]){
			this._itemNumPropName += "_";
		}

		// Step 4: Some data files specify an optional 'identifier', which is 
		// the name of an attribute that holds the identity of each item. 
		// If this data file specified an identifier attribute, then build a 
		// hash table of items keyed by the identity of the items.
		var arrayOfValues;

		var identifier = dataObject.identifier;
		if(identifier){
			this._itemsByIdentity = {};
			this._features['dojo.data.api.Identity'] = identifier;
			for(i = 0; i < this._arrayOfAllItems.length; ++i){
				item = this._arrayOfAllItems[i];
				
				arrayOfValues = item[identifier];
				var identity = arrayOfValues[0];
				if(!this._itemsByIdentity[identity]){
					this._itemsByIdentity[identity] = item;
				}else{
					if(this._jsonFileUrl){
						throw new Error("zero.ide.data.ideFilePathStore:  The json data as specified by: [" + this._jsonFileUrl + "] is malformed.  Items within the list have identifier: [" + identifier + "].  Value collided: [" + identity + "]");
					}else if(this._jsonData){
						throw new Error("zero.ide.data.ideFilePathStore:  The json data provided by the creation arguments is malformed.  Items within the list have identifier: [" + identifier + "].  Value collided: [" + identity + "]");
					}
				}
			}
		}else{
			this._features['dojo.data.api.Identity'] = Number;
		}
		
		// Step 5: Walk through all the items, and set each item's properties 
		// for _storeRefPropName and _itemNumPropName, so that store.isItem() will return true.
		for(i = 0; i < this._arrayOfAllItems.length; ++i){
			item = this._arrayOfAllItems[i];
			item[this._storeRefPropName] = this;
			item[this._itemNumPropName] = i;			
			
			var s = item["path"][0]; 
			if (s.length > this.resourceString.length) {
				item["path"][0] = s.substring(this.resourceString.length,s.length);
			} else if (s.length == this.resourceString.length) {
				item["path"][0] = '/';
			}
		}

		// Step 6: We walk through all the attribute values of all the items,
		// looking for type/value literals and item-references.
		//
		// We replace item-references with pointers to items.  For example, we change:
		//		{ name:['Kermit'], friends:[{_reference:{name:'Miss Piggy'}}] }
		// into this:
		//		{ name:['Kermit'], friends:[miss_piggy] } 
		// (where miss_piggy is the object representing the 'Miss Piggy' item).
		//
		// We replace type/value pairs with typed-literals.  For example, we change:
		//		{ name:['Nelson Mandela'], born:[{_type:'Date', _value:'July 18, 1918'}] }
		// into this:
		//		{ name:['Kermit'], born:(new Date('July 18, 1918')) } 
		//
		// We also generate the associate map for all items for the O(1) isItem function.
		for(i = 0; i < this._arrayOfAllItems.length; ++i){
			item = this._arrayOfAllItems[i]; // example: { name:['Kermit'], friends:[{_reference:{name:'Miss Piggy'}}] }
			for(key in item){
				arrayOfValues = item[key]; // example: [{_reference:{name:'Miss Piggy'}}]
				for(var j = 0; j < arrayOfValues.length; ++j) {
					value = arrayOfValues[j]; // example: {_reference:{name:'Miss Piggy'}}
					if(value !== null && typeof value == "object"){
						if(value._type && value._value){
							var type = value._type; // examples: 'Date', 'Color', or 'ComplexNumber'
							var mappingObj = this._datatypeMap[type]; // examples: Date, dojo.Color, foo.math.ComplexNumber, {type: dojo.Color, deserialize(value){ return new dojo.Color(value)}}
							if(!mappingObj){ 
								throw new Error("zero.ide.data.ideFilePathStore: in the typeMap constructor arg, no object class was specified for the datatype '" + type + "'");
							}else if(dojo.isFunction(mappingObj)){
								arrayOfValues[j] = new mappingObj(value._value);
							}else if(dojo.isFunction(mappingObj.deserialize)){
								arrayOfValues[j] = mappingObj.deserialize(value._value);
							}else{
								throw new Error("zero.ide.data.ideFilePathStore: Value provided in typeMap was neither a constructor, nor a an object with a deserialize function");
							}
						}
						if(value._reference){
							var referenceDescription = value._reference; // example: {name:'Miss Piggy'}
							if(dojo.isString(referenceDescription)){
								// example: 'Miss Piggy'
								// from an item like: { name:['Kermit'], friends:[{_reference:'Miss Piggy'}]}
								arrayOfValues[j] = this._itemsByIdentity[referenceDescription];
							}else{
								// example: {name:'Miss Piggy'}
								// from an item like: { name:['Kermit'], friends:[{_reference:{name:'Miss Piggy'}}] }
								for(var k = 0; k < this._arrayOfAllItems.length; ++k){
									var candidateItem = this._arrayOfAllItems[k];
									var found = true;
									for(var refKey in referenceDescription){
										if(candidateItem[refKey] != referenceDescription[refKey]){ 
											found = false; 
										}
									}
									if(found){ 
										arrayOfValues[j] = candidateItem; 
									}
								}
							}
						}
					}
				}
			}
		}
	},

	getIdentity: function(item){
      //if(zlog.isDebug())console.debug("ideFilePathStore getIdentity called ");		
     
		var identifier = this._features['dojo.data.api.Identity'];
		if(identifier === Number){
			return item[this._itemNumPropName]; // Number
		}else{
			var arrayOfValues = item[identifier];
			if(arrayOfValues){
				return arrayOfValues[0]; // Object || String
			}
		}
		return null; // null
	},

	fetchItemByIdentity: function(keywordArgs){
      //if(zlog.isDebug())console.debug("ideFilePathStore fetchItemByIdentity called ");		

		// Hasn't loaded yet, we have to trigger the load.
		if(!this._loadFinished){
			var self = this;
			if(this._jsonFileUrl){

				if(this._loadInProgress){
					this._queuedFetches.push({args: keywordArgs});
				}else{
					this._loadInProgress = true;
					var getArgs = {
							url: self._jsonFileUrl, 
							handleAs: "json-comment-optional",
							load: function(data){
								var scope =  keywordArgs.scope?keywordArgs.scope:dojo.global;
								try{
									self._getItemsFromLoadedData(data);
									self._loadFinished = true;
									self._loadInProgress = false;
									var item = self._getItemByIdentity(keywordArgs.identity);
									if(keywordArgs.onItem){
										keywordArgs.onItem.call(scope, item);
									}
									self._handleQueuedFetches();
								}catch(error){
									self._loadInProgress = false;
									if(keywordArgs.onError){
										keywordArgs.onError.call(scope, error);
									}
								}
							},
							error: function(error){
								self._loadInProgress = false;
								if(keywordArgs.onError){
									var scope =  keywordArgs.scope?keywordArgs.scope:dojo.global;
									keywordArgs.onError.call(scope, error);
								}
							}
					};
					dojo.xhrGet(getArgs);
				}

			}else if(this._jsonData){
				// Passed in data, no need to xhr.
				self._getItemsFromLoadedData(self._jsonData);
				self._jsonData = null;
				self._loadFinished = true;
				var item = self._getItemByIdentity(keywordArgs.identity);
				if(keywordArgs.onItem){
					var scope =  keywordArgs.scope?keywordArgs.scope:dojo.global;
					keywordArgs.onItem.call(scope, item);
				}
			} 
		}else{
			// Already loaded.  We can just look it up and call back.
			var item = this._getItemByIdentity(keywordArgs.identity);
			if(keywordArgs.onItem){
				var scope =  keywordArgs.scope?keywordArgs.scope:dojo.global;
				keywordArgs.onItem.call(scope, item);
			}
		}
	},

	_getItemByIdentity: function(identity){
      //if(zlog.isDebug())console.debug("ideFilePathStore _getItemByIdentity called ");		

		var item = null;
		if(this._itemsByIdentity){
			item = this._itemsByIdentity[identity];
		}else{
			item = this._arrayOfAllItems[identity];
		}
		if(item === undefined){
			item = null;
		}
		return item; // Object
	},

	getIdentityAttributes: function(item){
      //if(zlog.isDebug())console.debug("ideFilePathStore getIdentityAttributes called ");		
		 
		var identifier = this._features['dojo.data.api.Identity'];
		if(identifier === Number){
			// If (identifier === Number) it means getIdentity() just returns
			// an integer item-number for each item.  The dojo.data.api.Identity
			// spec says we need to return null if the identity is not composed 
			// of attributes 
			return null; // null
		}else{
			return [identifier]; // Array
		}
	},
	
	_forceLoad: function(){
      //if(zlog.isDebug())console.debug("ideFilePathStore _forceLoad called ");		

		var self = this;
		if(this._jsonFileUrl){
				var getArgs = {
					url: self._jsonFileUrl, 
					handleAs: "json-comment-optional",
					sync: true,
					load: function(data){
						try{
							//Check to be sure there wasn't another load going on concurrently 
							//So we don't clobber data that comes in on it.  If there is a load going on
							//then do not save this data.  It will potentially clobber current data.
							//We mainly wanted to sync/wait here.
							//TODO:  Revisit the loading scheme of this store to improve multi-initial
							//request handling.
							if (self._loadInProgress !== true && !self._loadFinished) {
								self._getItemsFromLoadedData(data);
								self._loadFinished = true;
							}
						}catch(e){
							console.log(e);
							throw e;
						}
					},
			      error: function(error){throw error;}
				};
			dojo.xhrGet(getArgs);
		}else if(this._jsonData){
			self._getItemsFromLoadedData(self._jsonData);
			self._jsonData = null;
			self._loadFinished = true;
		} 
	}
});
//Mix in the simple fetch implementation to this class.
dojo.extend(zero.ide.data.ideFilePathStore,dojo.data.util.simpleFetch);

}
